unit cCiaBuffer;

{
   This component is designed to receive binary streams where the first 4 bytes
   is the length of the stream. The length can be in Motorola format
   or Intel format (For Motorola format set SwapLen := True).

   When a complete stream is received then OnReceived event is fired. There (and
   only there) you call RcvdLen to allocate memory and then Receive or
   just call ReceiveStr if you need the data in a string. The first 4 bytes
   containing the length are not included in the Receive or ReceiveStr.

   It is optimized for speed. It will allocate / grow memory when needed but
   will reuse the allocated memory. Only pointers are moving.

   Examples:

   // sending
   var
      Len: integer;
      ToSend: string;
   begin
      ToSend := 'Hello';
      Len := Length(ToSend);
      WSocket1.Send(@Len, 4);
      WSocket1.Send(@ToSend[1], Len);
   end;

   // there are 2 ways to feed the component with data
   // this is fast because you copy direcly into the components buffer
   procedure TForm1.WSocket1DataAvailable(Sender: TObject; Error: Word);
   var
      Count: integer;
   begin
      If Error <> 0 then
         Exit;

      with Sender as TWSocket do
         Count := Receive(CiaBuffer1.MemoryWrite, CiaBuffer1.MemFree);
      CiaBuffer1.BytesWritten := Count;
      if CiaBuffer1.MemFree < 512 then
         CiaBuffer1.Grow(2048);
   end;

   // this is slower because it requires an extra copy of the received bytes
   procedure TForm1.WSocket1DataAvailable(Sender: TObject; Error: Word);
   var
      Count: integer;
      Buffer: array[1..1024] of char;
   begin
      if Error <> 0 then
         Exit;

      with Sender as TWSocket do
         Count := Receive(@Buffer[1], High(Buffer));
      CiaBuffer1.Add(@Buffer[1], Count);
   end;

   // there are 3 ways how to retreive the received stream:
   // Copied in a string
   procedure TForm1.CiaBuffer1ferReceived(Sender: TObject);
   begin
      with Sender as TCiaBuffer do
         Memo1.Lines.Add(ReceiveStr);
   end;

   // Copied in a memoryblock (you get a pointer to it)
   procedure TForm1.CiaBuffer1ferReceived(Sender: TObject);
   var
      P: PChar;
   begin
      with Sender as TCiaBuffer do
      begin
         GetMem(P, RcvdLen);
         Receive(P);
         // here you do something with P
         FreeMem(P);
      end;
   end;

   // No copy at all, for example if you have to pass a pointer to some
   // other component that does a copy itself
   procedure TForm1.CiaBuffer1ferReceived(Sender: TObject);
   var
      P: PChar;
   begin
      with Sender as TCiaBuffer do
         P := MemoryRead;
      // here you do something with P
   end;
}

interface

uses Classes;

type
  TCiaBuffer = class(TComponent)
  private
    FLen: integer;      // only for internal use
    FRcvdLen: integer;  // only for the property
    FReadPtr: integer;
    FWritePtr: integer;
    FRcvd: pointer;
    FRcvdSize: integer;
    FInitialSize: integer;
    FSwapLen: boolean;
    FOnReceived: TNotifyEvent;
    procedure SetInitialSize(value: integer);
    function  GetPointerToWriteTo: pointer;
    function  GetPointerToReadFrom: pointer;
    function  GetFreeMem: integer;
    procedure GetLenOfPacket;
    procedure SetBytesWritten(Value: integer);
    procedure DataAdded;
    procedure DataRead;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure Add(Data: pointer; Len: integer);
    procedure Grow(Len: integer);
    procedure Receive(Buffer: pointer);
    function  ReceiveStr: string;
    property BytesWritten: integer write SetBytesWritten;
    property MemoryWrite: pointer read GetPointerToWriteTo;
    property MemoryRead: pointer read GetPointerToReadFrom;
    property MemFree: integer read GetFreeMem;
  published
    property InitialSize: integer read FInitialSize write SetInitialSize;
    property RcvdLen: integer read FRcvdLen;
    property SwapLen: boolean read FSwapLen write FSwapLen;
    property OnReceived: TNotifyEvent read FOnReceived write FOnReceived;
  end;

procedure Register;

implementation

//------------------------------------------------------------------------------
procedure Register;
begin
   RegisterComponents('Cia', [TCiaBuffer]);
end;

//------------------------------------------------------------------------------
function Swap4(n: Cardinal): Cardinal;
asm
   bswap eax
end;

//------------------------------------------------------------------------------
constructor TCiaBuffer.Create(AOwner: TComponent);
begin
   inherited;
   SetInitialSize(4096);
end;

//------------------------------------------------------------------------------
destructor TCiaBuffer.Destroy;
begin
   if Assigned(FRcvd) then
      FreeMem(FRcvd);
   inherited;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.SetInitialSize(value: integer);
begin
   if value <> FInitialSize then
   begin
      FInitialSize := Value;
      if (FRcvdSize < value) and not (csDesigning in ComponentState) then
         ReAllocMem(FRcvd, value + 4);
   end;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.Grow(Len: integer);
begin
   ReallocMem(FRcvd, FRcvdSize + Len + 4);
   inc(FRcvdSize, Len);
end;

//------------------------------------------------------------------------------
function TCiaBuffer.GetFreeMem: integer;
begin
   Result := FRcvdSize - FWritePtr;
end;

//------------------------------------------------------------------------------
function TCiaBuffer.GetPointerToWriteTo: pointer;
begin
   result := @(PChar(FRcvd)[FWritePtr]);
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.GetLenOfPacket;
begin
   if FWritePtr - FReadPtr >= 4 then
   begin
      Move(PChar(FRcvd)[FReadPtr], FLen, 4);
      if FSwapLen then
         FLen := Swap4(FLen);
      inc(FReadPtr, 4);
   end;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.SetBytesWritten(Value: integer);
begin
   if Value <= 0 then
      Exit;

   inc(FWritePtr, Value);
   DataAdded;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.Add(Data: pointer; Len: integer);
begin
   if Len <= 0 then
      Exit;
   if Len > (FRcvdSize - FWritePtr) then
      Grow(Len);

   Move(PChar(Data)[0], PChar(FRcvd)[FWritePtr], Len);
   inc(FWritePtr, Len);
   DataAdded;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.DataAdded;
begin
   if (FLen = 0) then
      GetLenOfPacket;

   while (FLen > 0) and ((FWritePtr - FReadPtr) >= FLen) and Assigned(FOnReceived) do
   begin
      FRcvdLen := FLen;
      FOnReceived(Self);
   end;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.DataRead;
begin
   inc(FReadPtr, FLen);
   FLen := 0;
   if FReadPtr = FWritePtr then
   begin
      FReadPtr  := 0;
      FWritePtr := 0;
   end else
      GetLenOfPacket;
end;

//------------------------------------------------------------------------------
function TCiaBuffer.GetPointerToReadFrom: pointer;
begin
   result := @(PChar(FRcvd)[FReadPtr]);
   DataRead;
end;

//------------------------------------------------------------------------------
procedure TCiaBuffer.Receive(Buffer: pointer);
begin
   Move(PChar(FRcvd)[FReadPtr], PChar(Buffer)[0], FLen);
   DataRead;
end;

//------------------------------------------------------------------------------
function TCiaBuffer.ReceiveStr: string;
begin
   SetLength(Result, FLen);
   Receive(@Result[1]);
end;

end.
